@@ -9,6 +9,8 @@ use Cache::Apt::Lookup;
 use Cache::Apt::Package;
 use Debian::Debhelper::Dh_Lib;
 use Text::Diff;
+use IO::Uncompress::Gunzip qw(gunzip $GunzipError) ;
+use Parse::DebControl;
 use Exporter;
 
 =pod
@@ -19,7 +21,7 @@ Emdebian::Tools - Support functions for emdebian-tools
 
 =head1 Copyright and Licence
 
- Copyright (C) 2006, 2007  Neil Williams <codehelp@debian.org>
+ Copyright (C) 2006-2008  Neil Williams <codehelp@debian.org>
 
  This package is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
@@ -38,22 +40,42 @@ Emdebian::Tools - Support functions for emdebian-tools
 
 use vars qw(@ISA @EXPORT);
 @ISA=qw(Exporter);
-@EXPORT=qw(emdeb_versionstring extract_emdebversion
+@EXPORT=qw(emdeb_versionstring extract_emdebversion bug_config
 check_emdebian_control create_patches get_dpkg_cross_dir
 check_toolchains find_latest_gcc find_latest_libc get_targetsuite
 check_dist host_arch get_username get_workdir check_workdir
-prepare_checklist get_aptagent tools_version prepare_pristine );
+prepare_checklist get_aptagent tools_version prepare_pristine
+parse_xcontrol lookup_sourcepkg target_config lookup_dependencies
+get_standards_version get_primary write_target_config );
 
-my $OURVERSION = "0.7.0";
+my $OURVERSION = "1.0.0";
+
+my @extradirs=qw/rules.d config xsfbs patches-applied debhelper.in /;
 
 =head1 tools_version
 
 all scripts (even bash ones) call this function to get the current
 version string for the tools.
 
+Parses the Debian changelog file and retrieves the most recent
+version.
+
 =cut
 
 sub tools_version {
+	my $input = "/usr/share/doc/libemdebian-tools-perl/changelog.gz";
+	my $z = new IO::Uncompress::Gunzip $input
+		or carp "IO::Uncompress::Gunzip failed: $GunzipError\n";
+	my @output = <$z>;
+	my @list=();
+	foreach my $str (@output)
+	{
+		if ($str =~ /^emdebian-tools\s\((.*)\)\s[a-z]*;.*\n$/)
+		{
+			push @list, $1;
+		}
+	}
+	return $list[0] if @list;
 	return $OURVERSION;
 }
 
@@ -66,12 +88,38 @@ but available to all tools.
 
 sub get_config
 {
+	my $config;
 	my $dpkg_cross_dir = &get_dpkg_cross_dir;
+	# allow user-specific overrides
 	my $cfile = $dpkg_cross_dir . "/emsource";
 	# otherwise use the debconf default.
 	$cfile = "/etc/emsource.conf" if (! -f $cfile);
-	# read emsource config file.
-	my $config = Config::Auto::parse("$cfile", format => "colon");
+	if ( -f $cfile)
+	{
+		# read emsource config file, if any.
+		$config = Config::Auto::parse("$cfile", format => "colon");
+	}
+	# collect emsandbox config into the same hash
+	$cfile = $dpkg_cross_dir . "/emsandbox";
+	$cfile = "/etc/emsandbox.conf" if (! -f $cfile);
+	if ( -f $cfile)
+	{
+		my $sandbox = Config::Auto::parse("$cfile", format => "colon");
+		$config->{'workingdir'} = $sandbox->{'workingdir'};
+		# only override if tools not installed
+		if (not defined $config->{'targetsuite'})
+		{
+			$config->{'targetsuite'} = $sandbox->{'targetsuite'};
+		}
+		# if tools is installed and suite differs, carp.
+		# shouldn't really happen.
+		elsif ($config->{'targetsuite'} ne $sandbox->{'targetsuite'})
+		{
+			carp ("Conflict between targetsuite setting of emdebian-tools".
+			" and emdebian-rootfs! Use dpkg-reconfigure to set the same".
+			" value in both packages. Using emdebian-tools setting.");
+		}
+	}
 	return $config;
 }
 
@@ -121,6 +169,37 @@ sub get_aptagent
 	return $aptagent;
 }
 
+=head1 get_primary
+
+emdebian-tools needs to be able to query apt cache data from a Debian mirror
+that supports all cross-building architectures - these repositories are called
+'primary mirrors' in Debian. If /etc/apt/sources.list does not contain a primary
+Debian mirror, a primary mirror will have been configured in /etc/emsource.conf
+using debconf.
+
+get_primary returns the name of that mirror or undef if a primary already exists
+- to use with 'apt-cross -m', wrap the returned value in 'ftp://$val/debian'
+
+ my $val = &get_primary;
+ my $mirror = "";
+ $mirror = "-m ftp://$val/debian" if (defined $val);
+ system ("apt-cross $mirror -u");
+
+The default value is ftp.fr.debian.org so the default mirror to pass to
+apt-cross is:
+ apt-cross -v -m ftp://ftp.fr.debian.org/debian -u
+
+=cut
+
+sub get_primary
+{
+	my $primary;
+	my $config = &get_config();
+	$primary = $config->{'primary'};
+	return undef if ($primary eq "");
+	return $primary;
+}
+
 =head1 get_targetsuite
 
 Returns the debconf selection of the default target suite. Falls back
@@ -136,7 +215,7 @@ codenames: unstable, testing and stable.
 
 sub get_targetsuite
 {
-	my $suite  = "unstable";
+	my $suite  = "";
 	my $config = &get_config();
 	if (not defined($config->{'targetsuite'}))
 	{
@@ -145,25 +224,34 @@ sub get_targetsuite
 		carp ($msg);
 		$suite = &get_suite;
 	}
+	else
+	{
+		$suite = $config->{'targetsuite'};
+	}
 	&set_suite($suite);
 	return $suite;
 }
 
 =head1 check_workdir
 
-Ensures the new directory is writable by creating the directory ourselves,
-if that fails, return a message and let the script decide how to handle
-the error.
+Ensures the new directory is writable and tries to create it
+if it does not already exist. If that fails, return a message
+and let the script decide how to handle the error.
 
 =cut
 
 sub check_workdir
 {
 	my $workdir = $_[0];
+	my $msg = "Working directory '$workdir' exists but is not writable! Use\n";
+	$msg .= "'sudo dpkg-reconfigure emdebian-tools'\n to specify a writable location.\n";
+	$msg .= "See emsource(1) or emsandbox(1) for more information.\n";
+	# avoid dying within the module, let the script decide whether to die.
+	return $msg if ((! -w $workdir) and (-d $workdir));
 	system ("mkdir -p $workdir") if (! -d $workdir);
-	my $msg = "Working directory has been specified as '$workdir' but it does not exist\n";
+	$msg = "Working directory has been specified as '$workdir' but it does not exist\n";
 	$msg .= "and cannot be created using 'mkdir -p'!: $!\n";
-	$msg .= "See emsource(1) for more information.\n";
+	$msg .= "See emsource(1) or emsandbox(1) for more information.\n";
 	# avoid dying within the module, let the script decide whether to die.
 	return $msg if (! -d $workdir);
 	return "";
@@ -263,6 +351,8 @@ sub create_patches
 	{
 		# skip directories
 		next if ( -d "debian/$file");
+		# skip binary files (yes, really! - see sysvinit)
+		next if ( -B "debian/$file");
 		if (! -f "../$package.old/debian/xcontrol")
 		{
 			open (XCONTROL, ">../$package.old/debian/xcontrol")
@@ -270,8 +360,13 @@ sub create_patches
 			close (XCONTROL);
 		}
 		# ignore generated files.
-		next if ( $file =~ /\.orig$/);
-		next if ( $file =~ /\.rej$/);
+		if (( $file =~ /\.orig$/) or ( $file =~ /\.rej$/)
+			or ($file =~ /\.debhelper$/) or ($file =~ /\.substvars$/)
+			or ($file =~ /\.debhelper\.log$/) or ($file =~ /^files$/))
+		{
+			unlink ("../emdebian-$file.patch") if (-f "../emdebian-$file.patch");
+			next;
+		}
 		if ((! -f "../$package.old/debian/$file") && (! -d "../$package.old/debian/$file"))
 		{
 			# create a blank
@@ -289,10 +384,13 @@ sub create_patches
 			print PATCHFILE $diff;
 			close (PATCHFILE);
 		}
+		if (( -f "emdebian-$file.patch") and ($diff eq ''))
+		{
+			unlink ("emdebian-$file.patch");
+		}
 		chdir ("$cwd");
 	}
 	# handle changeovers to new code nicely.
-	my @extradirs = qw % rules.d control.in %;
 	foreach my $d (@extradirs)
 	{
 		if (( -d "debian/$d") && ( ! -d "../$package.old/debian/$d"))
@@ -324,31 +422,38 @@ sub create_patches
 					print PATCHFILE $diff;
 					close (PATCHFILE);
 				}
+				# instead remove the patch file if no diff.
+				else { unlink ("emdebian-$d-$file.patch"); }
 				chdir ("$cwd");
 			}
 		}
 	}
 	# copy new patches in debian/patches into new debian-patch files for SVN
 	# emsource copies in the opposite direction.
-	return if (! -d "debian/patches");
-	opendir (DEBIAN, "debian/patches") or croak ("Cannot open debian patches directory: $!");
+	my $patchdir;
+	$patchdir = "debian/patches-applied" if (-d "debian/patches-applied");
+	$patchdir = "debian/patches" if (-d "debian/patches");
+	return if (not defined $patchdir);
+	opendir (DEBIAN, $patchdir) or croak ("Cannot open debian patch directory: $!");
 	@patchfiles=grep(!/^\.\.?$/, readdir DEBIAN);
 	closedir (DEBIAN);
 	foreach my $file (@patchfiles)
 	{
+		next if (-d "$patchdir/$file");
 		# skip unchanged upstream patches
-		if (-f "../$package.old/debian/patches/$file")
+		if (-f "../$package.old/$patchdir/$file")
 		{
 			chdir ("../");
 			$diff = '';
-			$diff = diff "$package.old/debian/patches/$file", "$cwd/debian/patches/$file", { STYLE => "Unified" };
+			$diff = diff "$package.old/$patchdir/$file", 
+				"$cwd/$patchdir/$file", { STYLE => "Unified" };
 			chdir ("$cwd");
 			next if ($diff eq '');
 		}
 		# skip CDBS patch log files
 		next if ($file =~ /\.level-[0-9]\.log$/);
 		chdir ("../");
-		open (DPATCH, "$cwd/debian/patches/$file") or carp ("Cannot read source patch: $!\n");
+		open (DPATCH, "$cwd/$patchdir/$file") or carp ("Cannot read source patch: $!\n");
 		my @p=<DPATCH>;
 		close (DPATCH);
 		open (PATCH, ">debian-patch-$file") or carp ("Cannot open debian-patch-$file : $!\n");
@@ -408,6 +513,11 @@ sub prepare_checklist()
 	push @list, "libstdc++${libc_latest}-${gcc_latest}-pic-${arch}-cross";
 	push @list, "libgcc1-${arch}-cross";
 	push @list, &check_linux($arch);
+	my $multilib;
+	my $our_arch = `dpkg-architecture -qDEB_BUILD_ARCH`;
+	chomp ($our_arch);
+	$multilib = "libc6-dev-i386" if ($our_arch =~ /^amd64$/);
+	push @list, $multilib if (defined $multilib);
 	return \@list;
 }
 
@@ -469,6 +579,11 @@ Only use where the package NAME contains a mathematically correct
 version, e.g. gcc-3.4 vs gcc-4.1, libc6 vs libc7 etc.
 returns 0 if no gcc package can be found in the cache.
 
+Calling binlookup does slow things down a bit but it ensures that a
+real binary package exists, not just a metapackage. binlookup is only
+called if the iterator specifies a package name that matches the
+gcc regular expression.
+
 =cut
 
 sub find_latest_gcc()
@@ -484,7 +599,7 @@ sub find_latest_gcc()
 	my $iter = &get_cache_iter();
 	do {
 		$pkg = $iter->next;
-		$h{$pkg}++ if ($pkg);
+		$h{$pkg}++ if (($pkg) and ($pkg =~ /gcc-([0-9\.\-]+)$/) and (&binlookup($pkg)));
 	} while ($pkg);
 	my @list = sort keys (%h);
 	my $choice = 0;
@@ -503,6 +618,11 @@ sub find_latest_gcc()
 Only use where the package NAME contains a mathematically correct
 version, e.g. gcc-3.4 vs gcc-4.1, libc6 vs libc7 etc.
 
+Calling binlookup does slow things down a bit but it ensures that a
+real binary package exists, not just a metapackage. binlookup is only
+called if the iterator specifies a package name that matches the
+libc regular expression.
+
 =cut
 
 sub find_latest_libc()
@@ -518,7 +638,7 @@ sub find_latest_libc()
 	my $iter = &get_cache_iter();
 	do {
 		$pkg = $iter->next;
-		$h{$pkg}++ if ($pkg);
+		$h{$pkg}++ if (($pkg) and ($pkg =~ /libc([0-9]+)$/) and (&binlookup($pkg)));
 	} while ($pkg);
 	my @list = sort keys (%h);
 	my $choice = 0;
@@ -566,9 +686,13 @@ sub prepare_pristine
 	# may need to copy all directories, as long as this is a clean debian directory.
 	mkdir "../$package.old", 0755;
 	mkdir "../$package.old/debian", 0755;
-	mkdir "../$package.old/debian/rules.d", 0755 if (-d "debian/rules.d");
-	mkdir "../$package.old/debian/control.in", 0755 if (-d "debian/control.in");
+	foreach my $extra (@extradirs)
+	{
+		mkdir "../$package.old/debian/$extra", 0755 if (-d "debian/$extra");
+	}
 	mkdir "../$package.old/debian/patches", 0755;
+	mkdir "../$package.old/debian/patches-applied", 0755 
+		if (-d "debian/patches-applied");
 	# read all files from the debian/ directory.
 	opendir (DEBIAN, "debian") or croak ("Cannot open $cwd/debian directory: $!");
 	my @patchfiles=grep(!/^\.\.?$/, readdir DEBIAN);
@@ -587,43 +711,234 @@ sub prepare_pristine
 		chdir ("$cwd");
 	}
 	# back up existing files in debian/rules.d (if any) {gcc}
-	if ( -d "debian/rules.d")
+	# or config {busybox} or xsfbs {libx11}
+	foreach my $extra (@extradirs)
 	{
-		opendir (DEBIAN, "debian/rules.d") or die ("Cannot open debian rules.d directory: $!");
-		@patchfiles=grep(!/^\.\.?$/, readdir DEBIAN);
-		closedir (DEBIAN);
-		foreach my $file (@patchfiles)
+		if ( -d "debian/$extra")
 		{
-			# skip empty directories
-			next if (! -f "debian/rules.d/$file");
-			open (CPY, "debian/rules.d/$file") or die "Cannot open $file: $!";
-			@control=<CPY>;
-			close CPY;
-			chdir ("../$package.old/");
-			open (CPY, ">debian/rules.d/$file") or die "Cannot write $file: $!";
-			print CPY @control;
-			close CPY;
-			chdir ("$cwd");
+			opendir (DEBIAN, "debian/$extra") or 
+				die ("Cannot open debian $extra directory: $!");
+			@patchfiles=grep(!/^\.\.?$/, readdir DEBIAN);
+			closedir (DEBIAN);
+			foreach my $file (@patchfiles)
+			{
+				# skip empty directories
+				next if (! -f "debian/$extra/$file");
+				open (CPY, "debian/$extra/$file") or die "Cannot open $file: $!";
+				@control=<CPY>;
+				close CPY;
+				chdir ("../$package.old/");
+				open (CPY, ">debian/$extra/$file") or die "Cannot write $file: $!";
+				print CPY @control;
+				close CPY;
+				chdir ("$cwd");
+			}
 		}
 	}
-	return if (! -d "debian/patches");
+	my $patchdir;
+	$patchdir = "debian/patches-applied" if (-d "debian/patches-applied");
+	$patchdir = "debian/patches" if (-d "debian/patches");
+	return if (not defined $patchdir);
 	# now back up any existing files in debian/patches
-	opendir (DEBIAN, "debian/patches") or croak ("Cannot open debian patches directory: $!");
+	opendir (DEBIAN, $patchdir) or croak ("Cannot open debian patch directory: $!");
 	@patchfiles=grep(!/^\.\.?$/, readdir DEBIAN);
 	closedir (DEBIAN);
 	foreach my $file (@patchfiles)
 	{
 		# skip directories
-		next if (! -f "debian/patches/$file");
-		open (CPY, "debian/patches/$file") or die "Cannot open $file: $!";
+		next if (! -f "$patchdir/$file");
+		open (CPY, "$patchdir/$file") or die "Cannot open $file: $!";
 		@control=<CPY>;
 		close CPY;
 		chdir ("../$package.old/");
-		open (CPY, ">debian/patches/$file") or die "Cannot write $file: $!";
+		open (CPY, ">$patchdir/$file") or die "Cannot write $file: $!";
 		print CPY @control;
 		close CPY;
 		chdir ("$cwd");
 	}
 }
 
+=head1 bug_config
+
+Read the embug offline data file for details of cross-building bugs
+and return a hash, indexed by package name, containing the bug
+numbers if a bug has been filed.
+
+=cut
+
+sub bug_config
+{
+	my $apt_cross_dir = &get_dpkg_cross_dir;
+	my $cfile = $apt_cross_dir . "/embug";
+	if (not -f $cfile)
+	{
+		open (CFILE, ">$cfile") or
+			croak ("cannot create config file: $!\n");
+		close (CFILE);
+	}
+	my $config = Config::Auto::parse("$cfile", format => "colon");
+	return $config;
+}
+
+=head1 target_config
+
+Read the embug offline data file for details of cross-building patches
+that are out of date (i.e. do not apply successfully to the current
+Debian version) and return a hash, indexed by package name.
+
+=cut
+
+sub target_config
+{
+	my $apt_cross_dir = &get_dpkg_cross_dir;
+	my $cfile = $apt_cross_dir . "/emtarget";
+	if (not -f $cfile)
+	{
+		open (CFILE, ">$cfile") or
+			croak ("cannot create config file: $!\n");
+		close (CFILE);
+	}
+	my $config = Config::Auto::parse("$cfile", format => "colon");
+	return $config;
+}
+
+=head1 parse_xcontrol
+
+Returns a hash of the contents of debian/xcontrol using
+the Parse::DebControl module. See Parse::DebControl (3)
+for more information on the hash contents.
+
+Returns undefined if no debian/xcontrol file exists.
+
+=cut
+
+sub parse_xcontrol
+{
+	my $parser = new Parse::DebControl;
+	my $xcontrol;
+	my $options;
+	return $xcontrol if (! -f "debian/xcontrol");
+	$xcontrol = $parser->parse_file('./debian/xcontrol', $options);
+	return $xcontrol;
+}
+
+=head1 lookup_sourcepkg
+
+Returns a hash where the key is the submitted package name and the
+value is the returned source package for that binary or the package
+name if already a source package.
+
+First parameter is the architecture to query (to cope with modified
+packages), second parameter is the name of the package.
+
+e.g.
+
+ use Data::Dumper;
+ use Emdebian::Tools;
+ $var = &lookup_sourcepkg("arm", "libqof1");
+ print Dumper($var);
+ $VAR1 = {
+         'libqof1' => 'qof'
+         };
+
+=cut
+
+sub lookup_sourcepkg
+{
+	&check_cache_arch(shift);
+	return undef unless (&get_suite);
+	&init_cache(1);
+	my $package = shift;
+	my $emp = AptCrossPackage->new();
+	$emp->Package($package);
+	$emp = &srclookup($package);
+	my $src = (defined $emp->{SourcePkg}) ? $emp->{SourcePkg} : $emp->{Package};
+	return {$package, $src};
+}
+
+=head1 lookup_dependencies
+
+Returns a space-separated string containing the names of
+dependencies of the requested package. (Pass each to
+&lookup_sourcepkg to get the appropriate source package to build
+the dependency.)
+
+First parameter is the architecture to query (to cope with modified
+packages), second parameter is the name of the package.
+
+e.g.
+
+ use Emdebian::Tools;
+ print &lookup_dependencies ("arm", "libqof1");
+
+ libc6 libgda3-3 libglib2.0-0
+
+=cut
+
+sub lookup_dependencies
+{
+	my $arch = shift;
+	&check_cache_arch($arch);
+	return undef unless (&get_suite);
+	&init_cache(1);
+	my $package = shift;
+	my $emp = AptCrossPackage->new();
+	$emp->Package($package);
+	$emp = &lookup_pkg($emp);
+	my @list=();
+	my $str;
+	my $depend = $$emp->Depends;
+	foreach my $d (@$depend)
+	{
+		my $n = $$d->Package;
+		my $p = new AptCrossPackage;
+		$p->Distribution(&get_suite);
+		$p->Architecture($arch);
+		$p->Package($n);
+		&lookup_pkg($p);
+		push @list, $p->Package;
+	}
+	return join(" ", @list);
+}
+
+=head1 get_standards_version
+
+Simple subroutine that returns the current Debian Standards Version
+for packages built using emdebian-tools. This will, usually, match the
+current Standards Version in Debian unstable - this routine provides
+a single location to update the version when necessary.
+
+It could be possible to automate this value using a query against
+the debian-policy package in Debian but that could lead to emdebian-tools
+building packages that do not actually conform to that version of Policy.
+Instead, the version is updated when the code conforms to the updated
+Policy.
+
+=cut
+
+sub get_standards_version
+{
+	return "3.8.0";
+}
+
+sub write_target_config
+{
+	my $config = shift;
+	if (not defined $config)
+	{
+		print "no defined config.\n";
+		return;
+	}
+	my $apt_cross_dir = &get_dpkg_cross_dir;
+	my $cfile = $apt_cross_dir . "/emtarget";
+	open (CFILE, ">$cfile") or
+		carp ("cannot write config file: $!\n");
+	foreach my $pkg (sort keys %$config)
+	{
+		# write $pkg and $b to CFILE
+		print CFILE "$pkg: " . $config->{$pkg} . "\n" if (defined $config->{$pkg});
+	}
+	close (CFILE);
+}
+
 1;
